/* dmxlarge.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: DMXLARGE.PAS */


/******************* LARGE-SIGNAL ANALYSIS PROCEDURES *******************/


#include <curses.h>
#include <panel.h>
#include "dmxutil.h"
#include "cmplxops.h"
#include "diodemx.h"
#include "dmxlarge.h"
#include "scifunc.h"
#include "fft.h"

extern double vd, vj, id, ij, c, dt, rs, zc, phi;
extern double I0;
extern double delta;
extern double ggamma;
extern double cj0;
extern complex Zin, zemblo[9];
extern double LOfreq, Idc, Vdc, Vdc_desired, pdc, LOeff, pwrdiss, pavail;
extern boolean LO_display;
extern double tolerance;

extern char s[80];

/******************* LARGE-SIGNAL ANALYSIS PROCEDURES *******************/

double idj(double v)
{
  if (v <= 0)
    return 0.0;
  if (v > 0 && v < phi)
    return (I0 * (exp(delta * v) - 1.0));
  /* v >= phi */
  return ((delta * (v - phi) + 1.0) * I0 * exp(delta * phi));
}


double cj(double v)
{
  double denom;

  if (v < 0.99 * phi) {
    denom = X_tothe_Y(1.0 - v / phi, ggamma);
    return (cj0 / denom);
  } else
    return (10.0 * cj0);
}


void Harmonics(double *vd_, double *id_)
{
  waveform vd, id;
  double mag, phase;
  complex cmplx, cmplx1;
  freqvector realv, imagv, reali, imagi, LoadPwr, DiodePwr;
  short j, dummy, Ypos;
  boolean NegSign[8];


  memcpy(vd, vd_, sizeof(waveform));
  memcpy(id, id_, sizeof(waveform));
  /* FFT  vd & id */
  FFT(vd, realv, imagv, 32, 5, 1);
  FFT(id, reali, imagi, 32, 5, 1);

  /* Calc power dissipated in the diode and load */
  for (j = 1; j <= 8; j++) {
    NegSign[j - 1] = false;
    cmplx.x = realv[j];
    cmplx.y = -imagv[j];
    cmplx1.x = reali[j];
    cmplx1.y = imagi[j];
    cmplx = CMULT(cmplx, cmplx1);
    DiodePwr[j] = 500.0 * cmplx.x;	/*milliwatts */
    if (DiodePwr[j] < 0)
      NegSign[j - 1] = true;
    DiodePwr[j] = 4.342945 * log(fabs(DiodePwr[j]));	/*dBm */
    COM_TO_POLAR(cmplx1, &mag, &phase);
    LoadPwr[j] = 500.0 * mag * mag * zemblo[j].x;
    LoadPwr[j] = 4.342945 * log(LoadPwr[j]);
  }

  /* display */
//  textcolor(White);
//  textbackground(Blue);
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  clear();
  move(1, 16);
  addstr("POWERS IN LARGE-SIGNAL HARMONICS");
  move(2, 16);
//  textcolor(LightCyan);
  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
  addstr("             (dBm)");
  move(4, 13);
//  textcolor(Yellow);
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);

  addstr("Harmonic Number   Power Dissipation in");
  move(5, 13);
  addstr("      N            Diode           Zs");
  move(6, 13);
//  textcolor(LightGray);
  attrset(COLOR_PAIR(COLOR_WHITE));

  addstr("=======================================");
//  textcolor(Yellow);
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
//  Ypos = wherey;
  getyx(stdscr, Ypos, dummy);
  for (j = 1; j <= 8; j++) {
    move(Ypos + j, 19);
    sprintf(s, "%d", j);
    addstr(s);
    move(Ypos + j, 30);
    if (NegSign[j - 1]) {
//      textcolor(White);
//      textbackground(Magenta);
      attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
      sprintf(s, "%7.3f*", DiodePwr[j]);
      addstr(s);
//      textcolor(Yellow);
//      textbackground(Blue);
      attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
    } else {
      sprintf(s, "%7.3f", DiodePwr[j]);
      addstr(s);
    }
    move(Ypos + j, 45);
    sprintf(s, "%7.3f", LoadPwr[j]);
    addstr(s);
  }
  move(20, 1);
//  textcolor(White);
//  textbackground(Magenta);
  attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
  addstr("* indicates that the diode is providing power at this harmonic");
//  textcolor(LightGray);
//  textbackground(Blue);
  refresh();
}


void LO_prams(double *vd_, double *id_, complex * Zin, double *pwrdiss, double *Idc, double *Vdc)
{
  waveform vd, id;
  complex cmplx, cmplx1;
  freqvector realv, imagv, reali, imagi;
  short j;


  memcpy(vd, vd_, sizeof(waveform));
  memcpy(id, id_, sizeof(waveform));
  /* Calculate diode power dissipation */
  *pwrdiss = 0.0;
  for (j = 0; j <= 31; j++)
    *pwrdiss += vd[j] * id[j];
  *pwrdiss /= 32.0;

  /* Calculate Zin */
  /* FFT  vd & id */
  FFT(vd, realv, imagv, 32, 5, 1);
  FFT(id, reali, imagi, 32, 5, 1);

  /* Form Zin */
  cmplx.x = realv[1];
  cmplx.y = imagv[1];
  cmplx1.x = reali[1];
  cmplx1.y = imagi[1];
  *Zin = CDIV(cmplx, cmplx1);

  /* Get DC current & voltage */
  *Vdc = realv[0] * 0.5;
  *Idc = reali[0] * 0.5;
}


void write_LO_prams(void)
{
  clear();
  move(2, 0);
//  textcolor(White);
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  addstr("                    LO PARAMETERS\n\n");
//  textcolor(LightCyan);
  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
  addstr(" LO Frequency:");
//  textcolor(Yellow);
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  sprintf(s, "%8.4f GHz\n\n", 1.0e-9 * LOfreq);
  addstr(s);
//  textcolor(LightCyan);
  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
  addstr(" Zs:   ");
//  textcolor(Yellow);
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  sprintf(s, " Real part=%7.2f     Imag part=%7.2f\n", zemblo[1].x, zemblo[1].y);
  addstr(s);
//  textcolor(LightCyan);
  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
  addstr(" Zin:   ");
//  textcolor(Yellow);
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  sprintf(s, "Real part=%7.2f     Imag part=%7.2f\n\n", Zin.x, Zin.y);
  addstr(s);
//  textcolor(LightCyan);
  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
  addstr(" DC Bias:");
//  textcolor(Yellow);
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  sprintf(s, " Idc, mA=%7.3f           Vdc=%8.3f\n\n", 1000.0 * Idc, Vdc);
  addstr(s);
//  textcolor(LightCyan);
  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
  addstr(" Power parameters:\n");
//  textcolor(Yellow);
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  pdc = Vdc * Idc;
  LOeff = (pwrdiss - pdc) / pavail;
  sprintf(s, " Diode RF power dissipation, mW=%7.2f\n", 1000.0 * (pwrdiss - pdc));
  addstr(s);
  sprintf(s, " Diode DC power dissipation, mW=%7.2f\n", 1000.0 * pdc);
  addstr(s);
  sprintf(s, " Available RF LO power, mW=%7.2f     dBm: %5.2f\n", 1000.0 * pavail, 10.0 * log(pavail) / log(10.0) + 30.0);
  addstr(s);
  sprintf(s, " LO Efficiency=%7.4f\n", LOeff);
  addstr(s);
//  textcolor(LightGray);
  attrset(COLOR_PAIR(COLOR_WHITE));
  refresh();
}


void bisec(double rszc, double tol, double vi, double *vj)
{
  /* Used to calc. diode waveforms when cj0=0, or to find init.
     conditions in Integrate_Diode_Equations */
  double vjlow = 0.0;
  double vjhigh, vjest, func;

  if (vi <= 0.0) {
    *vj = 2.0 * vi;
    /* vi<0 */
    return;
  }
  vjhigh = 2.00001 * vi;
  vjest = vi;
  do {
    *vj = vjest;
    func = 2.0 * vi - rszc * idj(vjest) - vjest;
    if (func < 0) {		/* vjest is too great */
      vjhigh = vjest;
      vjest = (vjhigh + vjlow) / 2.0;
    } else if (func > 0) {	/* vjest is too small */
      vjlow = vjest;
      vjest = (vjlow + vjhigh) / 2.0;
    } else
      *vj = vjest;
  } while (fabs(vjest - *vj) >= tol);
  *vj = vjest;
}


void Integrate_Diode_Equations(double *vi, double *vr, double *vd, double *vj, double *id, double *ij, double *c)
{
  short k, j, jplus1, jminus1;
  short intpoints = 32;
  short itnum, Ypos, dummy;
  double vjsav, vjint, cint, ijint, dvjint, dvjsav, dvjmid, dtint, dt24, rszc, vi_int, delta_vi;
  double init_cond_tol = 0.005, steady_state_tol = 0.005;
  waveform dvjdt;

  /*bailout value for setting initial cond's */
  /*bailout for steady-state conditions */
  /*number of interpolation points */
  dtint = dt / intpoints;	/*increase intpoints for better stability */
  rszc = rs + zc;

  /* Set initial conditions */
  bisec(rszc, init_cond_tol, vi[0], vj);

  if (cj0 != 0.0) {
    /* Integrate diode equations over an LO cycle */
    /* In-line code is used for speed */
    dt24 = dt / 24.0;
    dvjmid = 0.0;
    dvjdt[31] = 0.0;
    vj[31] = vj[0];
    ij[31] = idj(vj[31]);
    c[31] = cj(vj[31]);
    itnum = 0;

    do {
      itnum++;
      for (j = 0; j <= 31; j++) {	/* integrate thru one LO cycle */
	jminus1 = j - 1;
	if (jminus1 < 0)
	  jminus1 += 32;
	c[j] = cj(vj[j]);
	ij[j] = idj(vj[j]);
	dvjdt[j] = (2.0 * vi[j] - vj[j] - rszc * ij[j]) / (rszc * c[j]);
	jplus1 = j + 1;
	if (j == 31) {
	  jplus1 = 0;
	  vjsav = vj[0];
	}
	if (vj[j] <= 0.0) {	/*ij=0 and c(vj) dominates-Adams Bashforth */
	  vjint = vj[j] + dt24 * (23.0 * dvjdt[j] - 16.0 * dvjmid + 5.0 * dvjdt[jminus1]);
	  cint = cj(vjint);
	  vi_int = 0.5 * (vi[j] + vi[jplus1]);
	  dvjsav = dvjmid;
	  dvjmid = (2.0 * vi_int - vjint) / (rszc * cint);
	  vj[jplus1] = vjint + dt24 * (23.0 * dvjmid - 16.0 * dvjdt[j] + 5.0 * dvjsav);
	} else {
	  /*interpolate the interval between two points */
	  delta_vi = (vi[jplus1] - vi[j]) / intpoints;
	  vjint = vj[j];
	  dvjint = dvjdt[j];
	  vi_int = vi[j];
	  for (k = 1; k <= intpoints - 1; k++) {
	    vjint += dtint * dvjint;
	    ijint = idj(vjint);
	    cint = cj(vjint);
	    vi_int += delta_vi;
	    dvjint = (2.0 * vi_int - vjint - rszc * ijint) / (rszc * cint);
	    if (k == intpoints / 2)
	      dvjmid = dvjint;
	  }
	  vj[jplus1] = vjint + dtint * dvjint;
	}
      }
      /*must use Euler's method & many points */
      getyx(stdscr, Ypos, dummy);
      addstr(" Iteration no. ");
//      textcolor(White);
//      textbackground(LightMagenta);
      attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
      sprintf(s, "%2d", itnum);
      addstr(s);
//      textcolor(Yellow);
//      textbackground(Blue);
      attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
      addstr(" completed in diode eqn subroutine\n");
      move(Ypos, 0);
    } while (fabs(vj[0] - vjsav) >= steady_state_tol);
    /* Bail out when steady state is reached */
//    putchar('\n');
  } else {			/* cj0=0 */
    for (j = 0; j <= 31; j++) {
      bisec(rszc, 0.001, vi[j], &vj[j]);
      c[j] = cj0;
      ij[j] = idj(vj[j]);
      dvjdt[j] = 0.0;
    }
    fprintf(stderr, "c[%i] = %e ij[%i] = %e dvjdt[%i] = %e\n", j, c[j], j, ij[j], j, dvjdt[j]);
  }
  /* cj0=0 */

  /* Form reverse wave */
  for (j = 0; j <= 31; j++) {
    id[j] = ij[j] + c[j] * dvjdt[j];
    vd[j] = vj[j] + rs * id[j];
    vr[j] = vi[j] - id[j] * zc;
  }
}


void LARGE_SIG(double vlosource, double *vbias, double tolerance, double *vd, double *vj, double *id, double *ij, double *c)
{
  double err, errchange, ersav, phase;
  double min_change = 0.005;
  double Vberr;
  complex cmplx, cmplx1, cmplx2, vref_freq, new_vinc_freq;
  waveform prev_vinc, vinc_initial, vinc, vref;
  freqvector coscomp, sincomp;
  complex rho[9];
  short itnum, j, linenum;
  boolean converged;
  boolean first_shot = true, wantsabreak = false;
  char ch, ch1;
  double TEMP1, TEMP2;


  /* Form reflection coefficients, rho[j] */
  cmplx.x = zc;
  cmplx.y = 0.0;
  for (j = 0; j <= 8; j++) {
    cmplx1 = CSUB(zemblo[j], cmplx);
    cmplx2 = CADD(zemblo[j], cmplx);
    rho[j] = CDIV(cmplx1, cmplx2);
    /*bailout cond on RMS error for slow convergence */
  }

  nodelay(stdscr, TRUE);
  noecho();
  do {				/*bias adjust loop */
    clear();
    move(0, 0);
//    textcolor(White);
    attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
    addstr("                           LARGE-SIGNAL ANALYSIS\n");
    move(2, 39);
//    textcolor(LightCyan);
    attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
    sprintf(s, "Vdc =%7.4f\n", Vdc);
    addstr(s);
    move(3, 39);
    sprintf(s, "Idc =%5.2f mA\n", 1000.0 * Idc);
    addstr(s);
    refresh();

    itnum = 0;
    err = 0.0;
    ersav = 0.0;
    converged = false;


    /* Calculate initial voltage wave */
    cmplx = zemblo[0];
    cmplx1 = zemblo[1];
    phase = atan2(cmplx1.y, cmplx1.x + zc);
    for (j = 0; j <= 31; j++) {
      TEMP1 = zc + cmplx1.x;
      TEMP2 = cmplx1.y;
      vinc_initial[j] = *vbias * zc / (zc + cmplx.x) + vlosource * zc * sin(2.0 * M_PI * j / 32 - phase) / sqrt(TEMP1 * TEMP1 + TEMP2 * TEMP2);
      if (first_shot)
	vinc[j] = vinc_initial[j];
      prev_vinc[j] = 0.0;
    }

    /* Begin iterative procedure */
    do {
      itnum++;
      /* clear the bottom of the screen */
      for (linenum = 16; linenum <= 23; linenum++) {
	move(linenum, 0);
	clrtoeol();
      }
      move(16, 0);
//      textcolor(White);
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      sprintf(s, " Beginning large-signal iteration no: %d\n", itnum);
      addstr(s);
//      textcolor(Yellow);
      attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);

      /* Save old incident wave to use later, in determining convergence */
      for (j = 0; j <= 31; j++)
	prev_vinc[j] = vinc[j];

      /* Calculate reflected wave, vref */
      Integrate_Diode_Equations(vinc, vref, vd, vj, id, ij, c);
//      textcolor(Yellow);
      attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
      addstr(" Finished integrating the diode equations\n");

      /* Check for bailout request */
      if ((ch1 = getch()) != ERR) {
	for (linenum = 16; linenum <= 23; linenum++) {
	  move(linenum, 0);
	  clrtoeol();
	}
	move(16, 0);
//      textcolor(White);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" **** Abort large-signal analysis? (y/n): ");
	refresh();
	/* wait for user response */
	nodelay(stdscr, FALSE);
	ch1 = getch();
	nodelay(stdscr, TRUE);
//      printf("%c\n", ch1);
	CAP(&ch1);
//      textcolor(Yellow);
	attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	if (ch1 != 'Y')
	  wantsabreak = false;
	else
	  wantsabreak = true;
      }

      /* FFT to get freq-domain reflected wave */
      FFT(vref, coscomp, sincomp, 32, 5, 1);	/* FFT to freq domain */
      addstr(" Through FFT\n");
      refresh();

      /* Multiply by reflection coefficients, rho, to get new freq-domain
         incident wave */
      for (j = 0; j <= 8; j++) {
	vref_freq.x = coscomp[j];
	vref_freq.y = sincomp[j];
	new_vinc_freq = CMULT(rho[j], vref_freq);
	coscomp[j] = new_vinc_freq.x;
	sincomp[j] = new_vinc_freq.y;	/* new inc wave in freq domain */
      }

      /* FFT to get time representation */
      FFT(vinc, coscomp, sincomp, 32, 5, 2);	/* FFT to time domain */
      addstr(" Through inverse FFT\n");
      refresh();

//      delay(1000);
//      sleep(1);
      /* Add new waveform to original */
      for (j = 0; j <= 31; j++)
	vinc[j] += vinc_initial[j];

      /* Calculate error function */
      ersav = err;
      err = 0.0;
      for (j = 0; j <= 31; j++) {
	TEMP1 = prev_vinc[j] - vinc[j];
	err += TEMP1 * TEMP1;
      }
      err = sqrt(err / 32.0);
      errchange = fabs(ersav - err);
/* p2c: dmxlarge.pas, line 442:
 * Note: Using % for possibly-negative arguments [317] */
      move(itnum % 14 + 1, 0);
//      textcolor(LightRed);
      sprintf(s, " %d  RMS error = %10.6f", itnum, err);
      addstr(s);
      refresh();

//      textcolor(Yellow);
      attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
      converged = (errchange < min_change || err < tolerance);
      if (wantsabreak) {
	itnum = 3;
	converged = true;
      }
    } while (!(converged && itnum > 2));
    LO_prams(vd, id, &Zin, &pwrdiss, &Idc, &Vdc);
    if (LO_display) {
      write_LO_prams();
//      delay(2500);
//      sleep(2);
      if ((ch = getch()) != ERR) {
//      textcolor(White);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr("\n Press any key to continue\n");
//      textcolor(Yellow);
	attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	ch = getch();
      }
    }

    /* modify vbias to get desired diode bias */
    Vberr = Vdc - Vdc_desired;
    if (fabs(Vberr) > 0.005)
      *vbias -= Vberr;
    first_shot = false;
  } while (!(fabs(Vberr) < 0.005 || wantsabreak));

//  textcolor(LightGray);
  attrset(COLOR_PAIR(COLOR_WHITE));
  nodelay(stdscr, TRUE);
  echo();
}


void do_lgsg(double vlosource, double *vbias)
{
  Vdc = 0.0;
  Idc = 0.0;
  LARGE_SIG(vlosource, vbias, tolerance, &vd, &vj, &id, &ij, &c);
}
