/* graphit.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: GRAPHIT.PAS */

#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include "diodemx.h"


void drawgraph(graphtype graph, double *wave)
{
  /* Graphing subroutine */
  double f[32];
  short fdots[32];
  char title[41], MaxStr[41], MinStr[41];
  double interval, max, min, dots_per_unit, dots_per_step, steps, scalecoef;
  short yposition, leftend, rightend, j;
  short FORLIM;
  int res;

  if (graph == junccap) {
    scalecoef = 1.0e12;
    strcpy(title, "JUNCTION CAP., pF");
  } else if (graph == diodev) {
    scalecoef = 1.0;
    strcpy(title, "DIODE VOLTAGE, VOLTS");
  } else if (graph == juncv) {
    scalecoef = 1.0;
    strcpy(title, "JUNCTION VOLTAGE,VOLTS");
  } else if (graph == diodecur) {
    scalecoef = 1000.0;
    strcpy(title, "DIODE CURRENT, mA");
  } else if (graph == junccur) {
    scalecoef = 1000.0;
    strcpy(title, "JUNCTION CURRENT, mA");
  } else if (graph == capcur) {
    scalecoef = 1000.0;
    strcpy(title, "CAPACITIVE CURRENT, mA");
  }

  max = scalecoef * wave[0];
  min = scalecoef * wave[0];

  /*find max/min of waveform */
  for (j = 0; j <= 31; j++) {
    f[j] = scalecoef * wave[j];
    if (f[j] > max)
      max = f[j];
    if (f[j] < min)
      min = f[j];
  }

  /*decide on max/min and interval for the vertical axis */
  if (fabs(max - min) < 1.2) {
    max = 0.1 * ((long) (10.0 * max) + 1.0);
    min = 0.1 * ((long) (10.0 * min) - 1.0);
    interval = 0.1;
  } else if (1.2 <= fabs(max - min) && fabs(max - min) <= 12.0) {
    max = (long) max + 1.0;
    min = (long) min - 1.0;
    interval = 1.0;
  } else {
    max = 10.0 * ((long) (0.1 * max) + 1.0);
    min = 10.0 * ((long) (0.1 * min) - 1.0);
    interval = 10.0;
  }
  if (graph == junccap && min < 0.0)
    min = 0.0;

  /*convert function values from reals to dots */
  /*there are 140 dots in the vertical scale */
  dots_per_unit = 140.0 / fabs(max - min);
  for (j = 0; j <= 31; j++)
    fdots[j] = 165 - (long) floor(dots_per_unit * (f[j] - min) + 0.5);

  res = vga_init();
  if (res < 0) {
    printf("Problem initializing graphics; we shall now crash.....\n");
    exit(-1);
  }

  res = vga_setmode(G320x200x256);

  if (res < 0) {
    printf("Problem with graphics driver; we shall now crash.....\n");
    exit(-1);
  }

  gl_setcontextvga(G320x200x256);
  gl_setfont(8, 8, gl_font8x8);
  gl_setwritemode(FONT_COMPRESSED + WRITEMODE_OVERWRITE);
  gl_setfontcolors(0, vga_white());
  vga_clear();

  fprintf(stderr, "after vga_clear()\n");

  sprintf(MaxStr, "%6.2f", max);
/* p2c: graphit.pas, line 106:
 * Note: Possible string truncation in assignment [145] */
  sprintf(MinStr, "%6.2f", min);
/* p2c: graphit.pas, line 107:
 * Note: Possible string truncation in assignment [145] */

  gl_line(76, 25, 300, 25, 1);
  gl_line(76, 25, 76, 165, 1);
  gl_line(76, 165, 300, 165, 1);
  gl_line(300, 165, 300, 25, 1);

  gl_printf(140, 176, "ONE LO CYCLE");
  gl_printf(108, 185, "(11.25 degrees/step)");
  gl_printf((224 - strlen(title) * 8) / 2 + 76, 16, title);
  gl_printf(22, 165, MinStr);
  gl_printf(22, 33, MaxStr);

  fprintf(stderr, "before draw graph\n");

  /*draw graph */
  for (j = 1; j <= 31; j++) {
    leftend = j * 7 + 69;
    rightend = leftend + 7;
    gl_line(leftend, fdots[j - 1], rightend, fdots[j], 1);	/*draw graph */
    if ((j & 3) == 0)		/*major time-axis tick */
      gl_line(rightend, 165, rightend, 159, 1);
    else
      gl_line(rightend, 165, rightend, 162, 1);
  }
  /*minor time axis ticks */
  gl_line(293, fdots[31], 300, fdots[0], 1);	/*last element of graph */

  /*ticks on vertical scale */
  if (fabs(max - interval) < 0.95 * interval && fabs(min + interval) < 0.95 * interval)
    /*i.e., if both max and min */
    interval = 0.1 * interval;
  /*  are just one interval  */
  steps = (max - min) / interval;
  dots_per_step = 140.0 / steps;
  FORLIM = (long) floor(steps - 1 + 0.5);
  for (j = 1; j <= FORLIM; j++) {
    yposition = (long) floor(25 + j * dots_per_step + 0.5);
    if (fabs(max - j * interval) < 0.95 * interval) {
      gl_line(76, yposition, 300, yposition, 1);	/*write zero line */
      gl_printf(22, yposition + 4, "  0.00");
    } else
      gl_line(76, yposition, 79, yposition, 1);
  }
  /* wait for a key */
  vga_getch();

  vga_setmode(TEXT);

}
