/* noise.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: NOISE.PAS */

#include "diodemx.h"
#include "cmplxops.h"
#include "cinvert.h"
#include "scifunc.h"
#include "fft.h"
#include "noise.h"

extern double temp, rs;
extern complex one, zero;

void NOISE(cmplx_mtx JYconv_, Zsigtype Zembsig_, complex Zinmatch_, complex Zoutmatch_, complex Zout_, short inpfreq, short outfreq, double conv_loss, double MAGain, waveform ij_, boolean conjmatch, double *TSSB, double *Tout)
{

  complex IJfreq[17];
  complex Ctherm[9], YeRs[9];
  cmplx_mtx Corrmtx, Zaug;
  freqvector IJreal, IJimag;
  complex cmplx, cmplx1, Vnjunc;
  double ZemRs[9];
  short j, k;
  /*set physical constants */
  double electron = 1.609e-19, Boltz = 1.37e-23;
  double Zem, phase, gain, num, VnMagSqr, Pnoise;
  boolean stable;

  cmplx_mtx JYconv;
  Zsigtype Zembsig;
  complex Zinmatch, Zoutmatch, Zout;
  waveform ij;

  memcpy(JYconv, JYconv_, sizeof(cmplx_mtx));
  memcpy(Zembsig, Zembsig_, sizeof(Zsigtype));
  Zinmatch = Zinmatch_;
  Zoutmatch = Zoutmatch_;
  Zout = Zout_;
  memcpy(ij, ij_, sizeof(waveform));


  /*Set Zemb=Zin/Zout if conj match data is desired */
  stable = !(conjmatch && Zinmatch.y == 0.0 && Zoutmatch.y == 0.0);
  if (!stable) {
    *Tout = 0.0;
    *TSSB = 0.0;
    goto _L625;
  }
  if (conjmatch) {
    if (inpfreq < 0)
      Zinmatch = CONJ(Zinmatch);
    if (outfreq < 0)
      Zoutmatch = CONJ(Zoutmatch);
    Zinmatch = CONJ(Zinmatch);
    Zoutmatch = CONJ(Zoutmatch);
    Zembsig[inpfreq + 4] = Zinmatch;
    Zembsig[outfreq + 4] = Zoutmatch;
    Zout = Zoutmatch;
    conv_loss = MAGain;
  }

  /*FFT ij and find shot noise correlation matrix elements */
  FFT(ij, IJreal, IJimag, 32, 5, 1);
  cmplx.x = electron * IJreal[0];
  cmplx.y = 0.0;
  IJfreq[8] = cmplx;
  for (j = 1; j <= 8; j++) {
    cmplx.x = electron * IJreal[j];
    cmplx.y = electron * IJimag[j];
    IJfreq[j + 8] = cmplx;
    cmplx.y = -cmplx.y;
    IJfreq[8 - j] = cmplx;
  }

  /*Form thermal noise correlation matrix elements and conversion matrix
     augmentation elements */
  cmplx.y = 0.0;
  num = 4.0 * temp * Boltz * rs;
  for (j = 0; j <= 8; j++) {
    cmplx1 = Zembsig[j];
    cmplx1.x += rs;
    YeRs[j] = CDIV(one, cmplx1);
    COM_TO_POLAR(cmplx1, &Zem, &phase);
    ZemRs[j] = Zem * Zem;
    cmplx.x = num / ZemRs[j];
    Ctherm[j] = cmplx;
  }
  COM_TO_POLAR(Zout, &Zem, &phase);
  cmplx.x = num / (Zem * Zem);
  Ctherm[outfreq + 4] = cmplx;

  /*Form correlation and conversion matrices */
  for (j = 0; j <= 8; j++) {
    for (k = 0; k <= 8; k++) {
      if (j == k) {
	JYconv[j][k] = CADD(JYconv[j][k], YeRs[j]);
	Corrmtx[j][k] = CADD(Ctherm[j], IJfreq[8]);
      } else
	Corrmtx[j][k] = IJfreq[j - k + 8];
    }
  }
  Cinvert(JYconv, Zaug);

  /*Evaluate quadratic form to get noise temp */
  Vnjunc = zero;
  for (j = 0; j <= 8; j++) {
    cmplx1 = zero;
    for (k = 0; k <= 8; k++) {
      cmplx = Zaug[outfreq + 4][k];
      cmplx.y = -cmplx.y;
      cmplx = CMULT(Corrmtx[j][k], cmplx);
      cmplx1 = CADD(cmplx, cmplx1);
    }
    cmplx1 = CMULT(cmplx1, Zaug[outfreq + 4][j]);
    Vnjunc = CADD(cmplx1, Vnjunc);
  }

  /*Find output noise power spectral density and convert to temp */
  cmplx = Zembsig[outfreq + 4];
  COM_TO_POLAR(Vnjunc, &VnMagSqr, &phase);
  Pnoise = VnMagSqr * cmplx.x / ZemRs[outfreq + 4];
  gain = X_tothe_Y(10.0, conv_loss / -10.0);
  *Tout = Pnoise / Boltz;
  *TSSB = *Tout / gain;

_L625:;
}
